"""Convert JSON file to CSV file.

Adapted from https://stackoverflow.com/a/28246154

Does not handle more than one level of nesting
"""
import argparse
import csv
import json
from pathlib import Path


def flattenjson(b, delim):
    val = {}
    for i in b.keys():
        if isinstance(b[i], dict):
            get = flattenjson(b[i], delim)
            for j in get.keys():
                val[i + delim + j] = get[j]
        else:
            val[i] = b[i]

    return val


def main(args):
    json_file = Path(args.json_file)
    with json_file.open(mode="r") as fp:
        json_ = fp.read()
    json_obj = json.loads(json_)

    flat_json = list(map(lambda x: flattenjson(x, "__"), json_obj))

    columns = [x for row in flat_json for x in row.keys()]
    columns = list(set(columns))

    csv_file = json_file.with_suffix(".csv")

    with csv_file.open(mode="w", encoding="utf8") as out_file:
        csv_w = csv.writer(out_file)
        csv_w.writerow(columns)

        for i_r in flat_json:
            csv_w.writerow(map(lambda x: i_r.get(x, ""), columns))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Convert json to CSV file")
    parser.add_argument("json_file", action="store", help="JSON file to convert")

    args = parser.parse_args()

    main(args)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas.datastructures import Network

__all__ = ["ClayStructure"]


class ClayStructure(Network):
    def __init__(self, clay_bullets):
        super(ClayStructure, self).__init__()
        self._clay_bullets = clay_bullets
        self.network_from_clay_bullets(self._clay_bullets)
        self.update_default_edge_attributes(relation=None)
        self.update_default_edge_attributes(is_touching=False)

    @property
    def clay_bullets(self):
        return self.vertices

    @property
    def average_compressed_radius(self):
        sum_ = sum([bullet.compressed_radius for bullet in self._clay_bullets])
        return sum_ / len(self._clay_bullets)

    def _edges_from_distance(self, i, clay_bullet):
        edges = []
        for j, other_bullet in enumerate(self._clay_bullets):
            if i == j:
                continue
            dist = clay_bullet.plane.Origin.DistanceTo(other_bullet.plane.Origin)
            if dist <= clay_bullet.compressed_radius + other_bullet.compressed_radius:
                edges.append((i, j))  # equivalent to set.update()
        return edges

    def _set_attributes_edges_longer_than(self, dist, **kwargs):
        if len(kwargs) < 1:
            raise Exception("No attributes to set")

        keys = []
        for u, v in self.edges():
            if self.edge_length(u, v) >= dist:
                keys.append((u, v))

        self.set_edges_attributes(kwargs.keys(), kwargs.values(), keys=keys)

    def _bullet_neighboors_below(self, u):
        z_value = self.get_vertex_attribute(u, "z")
        bullets_below = self.vertices_where({"z": (0, z_value)})

        bullets_below_keys = [(u, v) for v in bullets_below if v != u]
        for u, v in bullets_below_keys:
            if self.edge_length(u, v) <= 20:
                self.add_edge(u, v, relation="neighboor_below", is_touching=True)

    def network_from_clay_bullets(self, clay_bullets):
        for i, clay_bullet in enumerate(clay_bullets):
            self.add_vertex(
                key=i,
                x=clay_bullet.plane.Origin.X,
                y=clay_bullet.plane.Origin.Y,
                z=clay_bullet.plane.Origin.Z,
                class_instance=clay_bullet,
            )

        edges_from_order = [(i, i + 1) for i in range(len(clay_bullets) - 1)]

        for u, v in edges_from_order:
            self.add_edge(u, v, relation="print_order", is_touching=True)

        # TODO: Better distance value
        self._set_attributes_edges_longer_than(26, is_touching=False)

        for i in range(len(self._clay_bullets)):
            self._bullet_neighboors_below(i)



"""This module manages communications and script execution.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import socket
import struct
import traceback

# ------ Wraps communications


def concatenate_script(list_ur_commands):
    """
    Internal function that concatenates generated UR script into one large
    script file. Usually used to combine scripts generated by the
    GrasshopperPython components

    Args:
        list_ur_commands: A list of formatted UR Script strings

    Returns:
        ur_script: The concatenated script
    """

    ur_script = "\ndef clay_script():\n"
    # ur_script += '\tpopup("running MAS_clay_shooting_script")\n'

    combined_script = ""
    for ur_cmd in list_ur_commands:
        combined_script += ur_cmd

    # format combined script
    lines = combined_script.split("\n")
    for l in lines:
        ur_script += "\t" + l + "\n"

    ur_script += "end\n"
    ur_script += "\nclay_script()\n"
    return ur_script


def stop_script():
    """
    Function that creates a UR script to stop both axis and ur robot

    Returns:
        ur_script: The stopping script
    """

    ur_script = "\ndef stop_script():\n"
    ur_script += '\tpopup("stopping script")\n'
    # Call axis to stop
    ur_script += "\tstop_ext_axis()\n"
    ur_script += "\twait_ext_axis()\n"
    # Call UR robot to stop
    ur_script += "\tstop program\n"
    ur_script += "\nend\n"
    ur_script += "\nstop_script()\n"
    return ur_script


def send_script(script_to_send, robot_id, offline_simulation):
    """
    Opens a socket to the Robot and sends a script

    Args:
        script_to_send: Script to send to socket
        robot_id: Integer. ID of robot

    """
    """Function that opens a socket connection to the robot"""
    PORT = 30002
    HOST = get_ip_ur(robot_id, offline_simulation)

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(2)
    try:
        s.connect((HOST, PORT))
    except:  # noqa E722
        print("Cannot connect to ", HOST, PORT)

    s.settimeout(None)
    max_size = 2 << 18
    n = len(script_to_send)
    if n > max_size:
        raise Exception("Program too long")

    try:
        s.send(script_to_send)
    except:  # noqa E722
        print("failed to send")
    s.close()


def get_ip_ur(ur_number, offline_simulation=False):
    """
    Function that gets the ip of the robot

    Args:
        ur_number: ID of robot (1,2 or 3)

    Returns:
        ip: string.
    """
    subnet = "192.168.10."
    if not offline_simulation:
        ip = subnet + str(ur_number + 9)
    else:
        ip = "localhost"
    print(ip)
    return ip


def _get_ip_axis(ur_number):
    """
    Internal function that gets the ip of the axis machine

    Args:
        ur_number: ID of robot (1,2 or 3)

    Returns:
        ip: string.
    """
    ip = 10 * ur_number
    return "192.168.10.%d" % ip


def create_ur_script(template, scripts, robot_id):
    """Function that creates final UR Script to be sent to the robot.

    It needs to be used with a template UR script file.
    The final script is used for a robot-axis setup.

    Args:
        template: A template UR script. This is usually fixed
        scripts: scripts to add to the body of the final script
        robot_id: ID of robot (1,2 or 3)

    Returns:
        ur_script: Formatted final scrip to send to robot
    """

    header = template.replace("<<<ip_axis>>>", _get_ip_axis(robot_id))
    body = concatenate_script(scripts)
    combined_script = header + body
    lines = combined_script.split("\n")

    ur_script = ""
    for l in lines:
        ur_script += "\t" + l + "\n"
    return ur_script


# ------ Real time


def listen_to_robot(robot_id, offline_simulation=False):
    PORT = 30003
    HOST = get_ip_ur(robot_id, offline_simulation)
    # Create dictionary to store data
    chunks = {}
    chunks["target_joints"] = []
    chunks["actual_joints"] = []
    chunks["forces"] = []
    chunks["pose"] = []
    chunks["time"] = [0]

    data = read(HOST, PORT)
    get_messages(data, chunks)
    return chunks


def read(HOST, PORT):
    """
    Method that opens a TCP socket to the robot, receives data from the robot
    server and then closes socket

    Returns:
        data: Data broadcast by the robot. In bytes
    """

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(0.1)
    try:
        s.connect((HOST, PORT))
        print("connected")
    except:  # noqa E722
        traceback.print_exc()
        print("Cannot connect to {}, {}".format(HOST, PORT))
    # s.settimeout(None)
    data = s.recv(1024)
    s.close()
    return data


def get_messages(bytes, chunks_info):
    """
    Function parses data stream and selects the following information:
    1) q_target
    2) q_actual
    3) TCP force
    4) Tool Vector
    5) Time

    This data is formatted and the chunks dictionary is updated
    for more info see: http://wiki03.lynero.net/Technical/RealTimeClientInterface
    """

    # get messages
    q_target = bytes[12:60]
    q_actual = bytes[252:300]
    tcp_force = bytes[540:588]
    tool_vector = bytes[588:636]
    controller_time = bytes[740:748]

    # format type: int,
    fmt_double6 = "!dddddd"
    fmt_double1 = "!d"

    # Unpack selected data
    target_joints = struct.unpack(fmt_double6, q_target)
    chunks_info["target_joints"] = (math.degrees(j) for j in target_joints)
    actual_joints = struct.unpack(fmt_double6, q_actual)
    chunks_info["actual_joints"] = (math.degrees(j) for j in actual_joints)
    forces = struct.unpack(fmt_double6, tcp_force)
    chunks_info["forces"] = forces
    pose = struct.unpack(fmt_double6, tool_vector)
    chunks_info["pose"] = pose
    time = struct.unpack(fmt_double1, controller_time)
    chunks_info["time"] = time



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import random

import compas.geometry as cg
from compas import IPY

if IPY:
    import Rhino.Geometry as rg
    from compas_rcf.utils.rhino_to_compas import rgplane_to_cgframe
    from compas_rcf.utils.rhino_to_compas import rgpoint_to_cgpoint


def rand_vector(dimensions):
    """Create a random vector.

    Pads vector if fewer than three dimensions up to three dimensions.

    Parameters
    ----------
    dimensions : int

    Returns
    -------
    `class`:compas.geometry.Vector
        Unitized random vector

    Notes
    -----
    Adapted from `Stack Overflow <https://stackoverflow.com/a/8453514>`_
    """
    v = [random.gauss(0, 1) for i in range(dimensions)]
    magnitude = sum(x ** 2 for x in v) ** 0.5
    vector = [x / magnitude for x in v]
    if len(vector) < 3:
        for i in range(3 - len(vector)):
            vector += [0]

    vector = cg.Vector(vector)

    vector.unitize()

    return vector


def remap_values(values, from_domain, to_domain, include_clipped=False):
    # TODO: Debug (reverse values compared to remap values in gh)
    from_min, from_max = from_domain
    to_min, to_max = to_domain

    from_range = from_max - from_min
    to_range = to_max - to_min

    remapped_values = []

    if not hasattr(values, "__iter__"):
        values = [values]

    for value in values:
        new_value = (((value - from_min) * to_range) / from_range) + to_min

        if to_min < new_value < to_max:
            remapped_values.append(new_value)

        elif include_clipped:
            remapped_values.append(new_value)

    if len(remapped_values) == 1:
        return remapped_values[0]
    else:
        return remapped_values


def flatten_list(l):
    return [item for sublist in l for item in sublist]


def shift_list(seq, shift=1):
    """Shift indices of list, wrapping at end.

    Notes
    -----
    Source: https://stackoverflow.com/a/29498813
    """
    return seq[-shift:] + seq[:-shift]


def list_elem_w_index_wrap(l, i):
    return l[i % len(l)]


def flip_matrix(listlike):
    """Rotate 2D-array.

    Parameters
    ----------
    listlike: iterable
        array to rotate

    Returns
    -------
    iterable
        rotated array

    Notes
    -----
    Taken from https://stackoverflow.com/a/496056
    """
    return zip(*listlike[::-1])


def ensure_frame(frame_like):
    if isinstance(frame_like, cg.Frame):
        return frame_like

    if isinstance(frame_like, cg.Plane):
        return cg.Frame.from_plane(frame_like)

    if IPY:
        if isinstance(frame_like, rg.Plane):
            return rgplane_to_cgframe(frame_like)
        if isinstance(frame_like, rg.Point3d):
            pt = rgpoint_to_cgpoint(frame_like)
            return cg.Frame(pt, cg.Vector(0, 1, 0), cg.Vector(1, 0, 0))

    raise TypeError(
        "Can't convert {} to compas.geometry.Frame".format(type(frame_like))
    )


def get_offset_frame(frame, distance):
    """Offset a frame in its Z axis direction.

    Parameters
    ----------
    frame : `class`:compas.geometry.Frame
        Frame to offset
    distance : float
        Translation distance in mm

    Returns
    -------
    `class`:compas.geometry.Frame
    """
    offset_vector = frame.zaxis * distance * -1
    T = cg.Translation(offset_vector)

    return frame.transformed(T)



# -*- coding: utf-8 -*-

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = "1.0"

from compas_rcf import __version__
import sphinx_compas_theme

# -- General configuration ------------------------------------------------

project = "compas_rcf"
copyright = "MAS DFAB 1920 students and tutors"
author = "Anton T Johansson"
version = release = __version__

master_doc = "index"

source_suffix = {".rst": "restructuredtext", ".md": "markdown"}
templates_path = [
    "_templates",
]
exclude_patterns: list = []

pygments_style = "sphinx"
show_authors = True
add_module_names = True
language = None

# -- Extension configuration ------------------------------------------------

extensions = [
    "recommonmark",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
]

# autodoc options
autodoc_default_options = {
    "member-order": "bysource",
    "special-members": "__init__",
    "exclude-members": "__weakref__",
    "undoc-members": True,
    "private-members": True,
    "show-inheritance": True,
}

autodoc_typehints = "signature"
autodoc_mock_imports = ["Rhino", "tkinter"]
autodoc_member_order = "alphabetical"

autoclass_content = "class"


# autosummary options

autosummary_generate = True

# napoleon options

napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = False
napoleon_use_rtype = False


# intersphinx options

intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "compas": (
        "https://compas-dev.github.io/main",
        "https://compas-dev.github.io/main/objects.inv",
    ),
    "compas_fab": (
        "https://gramaziokohler.github.io/compas_fab/latest/",
        "https://gramaziokohler.github.io/compas_fab/latest/objects.inv",
    ),
}

# -- Options for HTML output ----------------------------------------------

html_theme = "compaspkg"
html_theme_path = sphinx_compas_theme.get_html_theme_path()

html_theme_options = {
    "package_name": "compas_rcf",
    "package_title": project,
    "package_version": release,
}

html_context: dict = {}
html_static_path: list = []
html_extra_path = [".nojekyll", "CNAME"]
html_last_updated_fmt = ""
html_copy_source = False
html_show_sourcelink = True
html_add_permalinks = ""
html_experimental_html5_writer = True
html_compact_lists = True



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas.geometry import Frame

from compas_rcf.utils.util_funcs import ensure_frame

try:
    from typing import List
except ImportError:
    pass


def format_joint_positions(joint_values):
    jpos_fmt = "[" + ", ".join(["{:.4f}"] * 6) + "]"
    return jpos_fmt.format(*joint_values)


def format_pose(frame_like):
    frame = ensure_frame(frame_like)

    pose_data = [c / 1000.0 for c in frame.origin.data] + frame.axis_angle_vector()
    pose_fmt = "p[" + ", ".join(["{:.4f}"] * 6) + "]"
    return pose_fmt.format(*pose_data)


def format_urscript_cmd(func):
    # @wraps(func)
    def wrapper(*arg, **kwargs):
        cmd = "\t{}\n".format(func(*arg, **kwargs))
        return cmd

    return wrapper


# Motion
@format_urscript_cmd
def movel(frame_to, accel=1.2, vel=0.25, time=0, zone=0):
    # type: (Frame, float, float, float, float) -> str
    pose = format_pose(frame_to)
    return "movel({:s}, a={:.2f}, v={:2f}, t={:2f} r={:2f})".format(
        pose, accel, vel, time, zone
    )


@format_urscript_cmd
def movej(joint_positions, accel=1.4, vel=1.05, time=0, zone=0):
    # type: (List[float], float, float, float, float) -> str
    """
    Function that returns UR script for linear movement in joint space.

    Args:
        joints: A list of 6 joint angles (double).
        accel: tool accel in m/s^2
        accel: tool accel in m/s^2
        vel: tool speed in m/s

    Returns:
        script: UR script
    """
    # TODO: Test
    # TODO: Check acceleration and velocity are below set limit
    _joint_positions = format_joint_positions(joint_positions)

    return "movej({:s}, a={:.2f}, v={:.2f}), t={:.2f}, r={:.2f}".format(
        _joint_positions, accel, vel, time, zone
    )


# Utility


@format_urscript_cmd
def set_TCP(tcp_frame):
    # type: (Frame) -> str
    pose = format_pose(tcp_frame)
    return "set_tcp({:s})".format(pose)


@format_urscript_cmd
def textmsg(string):
    # type: (str) -> str
    return 'textmsg("' + string + '")'


@format_urscript_cmd
def popup(string):
    # type: (str) -> str
    # Popup title not implemented, neither is error or warning flags
    return 'popup("{}")'.format(string)


@format_urscript_cmd
def sleep(seconds):
    # type: (float) -> str
    """
    Function that returns UR script for sleep()

    Args:
        time: float.in s

    Returns:
        script: UR script
    """
    return "sleep({})\n".format(seconds)


# Communication


@format_urscript_cmd
def socket_open(server_address, server_port):
    # type: (str, int) -> str
    return 'socket_open("{}", {:d})'.format(server_address, server_port)


@format_urscript_cmd
def socket_send_string(text):
    # type: (str) -> str
    return 'socket_send_string("' + text + '")'


@format_urscript_cmd
def socket_close():
    # type: () -> str
    return "socket_close()"


# IO


@format_urscript_cmd
def set_DO(pin, state):
    # type: (int, bool) -> str
    # deprecation warning in UR manual
    # return "set_digital_out({:d}, {})".format(pin, state)
    return set_standard_digital_out(pin, state)


@format_urscript_cmd
def set_standard_digital_out(pin, state):
    # type: (int, bool) -> str
    """Set standard digital output signal level

    Parameters
    ----------
    pin : int, 0 to 7
        Pin number
    state: bool

    Returns
    -------
    str
    """
    # TODO: Test
    return "set_standard_digital_out({}, {})".format(id, str(state))


@format_urscript_cmd
def set_standard_analog_out(pin, signal_level):
    # type: (int, float) -> str
    """Set standard analog output level

    Parameters
    ----------
    pin : int, 0 to 1
        Pin number
    signal_level : float, 0 to 1
        Relative signal level

    Returns
    -------
    str
    """
    # TODO: Test

    return "set_standard_analog_out({}, {:.3})".format(id, signal_level)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_rhino import install

PACKAGES = ["compas_fab", "roslibpy", "compas_rcf", "compas_rrc"]


if __name__ == "__main__":

    packages = set(install.INSTALLABLE_PACKAGES + PACKAGES)

    install.install(version="6.0", packages=packages)



"""Will be replaced by urscript_wrapper.

This module wraps standard UR Script functions.
Main change is that plane information substitute for pose data
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import compas.geometry as cg
import Rhino.Geometry as rg
from compas.geometry.transformations import axis_angle_vector_from_matrix

from compas_rcf.ur.helpers import axis_angle_vector_from_plane_to_plane
from compas_rcf.utils.rhino_to_compas import rgtransform_to_matrix

# ----- UR Interfaces module -----


def set_analog_out(id, signal):
    """
    Function that returns UR script for setting analog out

    Args:
        id: int. Input id number
        signal: int. signal level 1(on) or 0(off)

    Returns:
        script: UR script
    """
    return "set_analog_out({:d}, {})\n".format(id, signal)


def set_digital_out(id, signal):
    """
    Function that returns UR script for setting digital out

    Args:
        id: int. Input id number
        signal: boolean. signal level - on or off

    Returns:
        script: UR script
    """

    # Format UR script
    return "set_digital_out({:d}, {})\n".format(id, signal)


def socket_open(address, port):
    """
    Function that returns UR script for setting digital out
    TODO(Jason) - some form of error checking?

    Args:
        address: string. IP address
        port: int. Port number

    Returns:
        script: UR script
    """

    script = "socket_open(%s,%s)\n" % (address, port)
    return script


def socket_send_string(ref_string):
    """
    Function that returns UR script for sending strings through a socket

    Args:
        ref_string: string. Either a string or name of variable defined in urscript

    Returns:
        script: UR script
    """

    script = "socket_send_string(%s)\n" % (ref_string)
    return script


# ----- UR Motion module -----

# Some Constants
MAX_ACCEL = 1.5
MAX_VELOCITY = 2


def _format_pose(pt_like, axis_angle):

    if isinstance(pt_like, (list, tuple)):
        pt_coords = [float(c) for c in pt_like]
        if len(pt_coords) < 3:
            for i in range(3 - len(pt_coords)):
                pt_coords += [0]
            # pt_coords += [0.] * (3 - len(pt_coords))
    elif isinstance(pt_like, rg.Point3d):
        pt_coords = [pt_like.X, pt_like.Y, pt_like.Z]
    elif isinstance(pt_like, cg.Point):
        pt_coords = pt_like.data
    elif isinstance(pt_like, rg.Plane):
        pt_coords = [pt_like.OriginX, pt_like.OriginY, pt_like.OriginZ]
    elif isinstance(pt_like, cg.Frame):
        pt_coords = pt_like.point.data
    else:
        raise TypeError("Could not convert argument to point")
    pose_data = [c / 1000.0 for c in pt_coords] + [float(a) for a in axis_angle]
    pose_fmt = "p[" + ", ".join(["{:.4f}"] * 6) + "]"
    return pose_fmt.format(*pose_data)


def _format_joint_positions(joint_values):
    jpos_fmt = "[" + ", ".join(["{:.4f}"] * 6) + "]"
    return jpos_fmt.format(*joint_values)


def move_l(plane_to, accel, vel, blend_radius=0):
    """
    Function that returns UR script for linear movement in tool-space.

    Args:
        plane_to: Rhino.Geometry Plane. A target plane for calculating pose
            (in UR base coordinate system)
        accel: tool accel in m/s^2
        vel: tool speed in m/s

    Returns:
        script: UR script
    """
    # Check acceleration and velocity are non-negative and below a set limit
    accel = MAX_ACCEL if (abs(accel) > MAX_ACCEL) else abs(accel)
    vel = MAX_VELOCITY if (abs(vel) > MAX_VELOCITY) else abs(vel)
    # Check blend radius is positive
    blend_radius = max(0, blend_radius)

    axis_angle_vector = axis_angle_vector_from_plane_to_plane(plane_to)

    # Create pose data
    pose = _format_pose(plane_to, axis_angle_vector)

    # Format UR script
    return "movel(%s, a = %.2f, v = %.2f, r = %.2f)\n" % (
        pose,
        accel,
        vel,
        blend_radius,
    )


def move_l_time(plane_to, time, blend_radius=0):
    """
    Function that returns UR script for linear movement in tool-space.

    Args:
        plane_to: Rhino.Geometry Plane. A target plane for calculating pose
            (in UR base coordinate system)
        time: Amount of time the movement should take, in seconds.
            Overrides speed and acceleration.

    Returns:
        script: UR script
    """
    # Check time is positive
    time = abs(time)
    # Check blend radius is positive
    blend_radius = max(0, blend_radius)

    axis_angle_vector = axis_angle_vector_from_plane_to_plane(plane_to)

    # Create pose data
    pose = _format_pose(plane_to, axis_angle_vector)

    # Format UR script
    return "movel(%s, a = %.2f, v = %.2f, t = %.2f, r = %.4f)\n" % (
        pose,
        1.2,
        0.25,
        time,
        blend_radius,
    )


def move_l2(plane_to, vel, blend_radius):
    # TODO: Test

    axis_angle_vector = axis_angle_vector_from_plane_to_plane(plane_to)

    # Create pose data
    pose = _format_pose(plane_to, axis_angle_vector)

    # Format UR script
    return "movel(%s, v = %.4f, r= %.4f)\n" % (pose, vel, blend_radius)


def move_j(joints, accel, vel):
    """
    Function that returns UR script for linear movement in joint space.

    Args:
        joints: A list of 6 joint angles (double).
        accel: tool accel in m/s^2
        accel: tool accel in m/s^2
        vel: tool speed in m/s

    Returns:
        script: UR script
    """
    # TODO: Test
    # TODO: Check acceleration and velocity are below set limit
    joint_positions = _format_joint_positions(joints)

    return "movej({}, a = {:.2f}, v = {:.2f})\n".format(
        joint_positions, abs(accel), abs(vel)
    )


def move_j_pose(plane_to, accel, vel, blend_radius=0.0):
    axis_angle_vector = axis_angle_vector_from_plane_to_plane(plane_to)

    # Create pose data
    pose = _format_pose(plane_to, axis_angle_vector)

    # Return UR script
    return "movej({}, a = {:.2f}, v = {:.2f}, r = {:.2f})\n".format(
        pose, accel, vel, blend_radius
    )


def move_c(plane_to, point_via, accel, vel):
    """
    Function that returns UR script for circular movement in tool-space.

    Only via planes, joint angles not wrapped

    Args:
        plane_to:  Rhino.Geometry Plane.A target plane used for calculating pose
            (in UR base coordinate system)
        point_via: Rhino.Geometry Point. A waypoint that movement passes through
        accel: tool accel in m/s^2
        vel: tool speed in m/s

    Returns:
        script: UR script
    """
    # TODO: Test
    # Check acceleration and velocity are non-negative and below a set limit
    accel = MAX_ACCEL if (abs(accel) > MAX_ACCEL) else abs(accel)
    vel = MAX_VELOCITY if (abs(vel) > MAX_VELOCITY) else abs(vel)

    axis_angle_vector = axis_angle_vector_from_plane_to_plane(plane_to)

    # Create pose data
    pose_to = _format_pose(plane_to, axis_angle_vector)
    pose_via = _format_pose(point_via, axis_angle_vector)

    # Format UR script
    return "movec({}, {}, a = {:.2f}, v = {:.2f})\n".format(
        pose_to, pose_via, accel, vel
    )


# ----- UR Internals module -----


def set_tcp_by_plane(x_offset, y_offset, z_offset, ref_plane=rg.Plane.WorldXY):
    """
    Function that returns UR script for setting tool center point

    Args:
        x_offset: float. tooltip offset in mm
        y_offset: float. tooltip offset in mm
        z_offset: float. tooltip offset in mm
        ref_plane: Plane that defines orientation of the tip. If none specified,
        world XY plane used as default. (in UR base coordinate system)

    Returns:
        script: UR script
    """
    # TODO: Test

    axis_angle_vector = axis_angle_vector_from_plane_to_plane(
        rg.Plane.WorldXY, plane_from=ref_plane
    )

    # Create pose data
    pose = _format_pose([x_offset, y_offset, z_offset, axis_angle_vector])

    # Return UR script
    return "set_tcp({})\n".format(pose)


def set_tcp_by_plane_angles(x_offset, y_offset, z_offset, x_rotate, y_rotate, z_rotate):
    """
    Function that returns UR script for setting tool center point

    Args:
        x_offset: float. tooltip offset in mm
        y_offset: float. tooltip offset in mm
        z_offset: float. tooltip offset in mm
        x_rotation: float. rotation around world x axis in radians
        y_rotation: float. rotation around world y axis in radians
        z_rotation: float. rotation around world z axis in radians

    Returns:
        script: UR script
    """
    # Create rotation matrix
    rX = rg.Transform.Rotation(x_rotate, rg.Vector3d(1, 0, 0), rg.Point3d(0, 0, 0))
    rY = rg.Transform.Rotation(y_rotate, rg.Vector3d(0, 1, 0), rg.Point3d(0, 0, 0))
    rZ = rg.Transform.Rotation(z_rotate, rg.Vector3d(0, 0, 1), rg.Point3d(0, 0, 0))
    R = rX * rY * rZ
    M = rgtransform_to_matrix(R)
    axis_angle_vector = axis_angle_vector_from_matrix(M)

    # Create pose data
    pose = _format_pose([x_offset, y_offset, z_offset], axis_angle_vector)

    # Format UR script
    return "set_tcp({})\n".format(pose)


def popup(message, title):
    """
    Function that returns UR script for popup

    Args:
        message: float. tooltip offset in mm
        title: float. tooltip offset in mm

    Returns:
        script: UR script
    """
    return 'popup("{}","{}") \n'.format(message, title)


def UR_log(message):
    return 'textmsg("{}")\n'.format(message)


def sleep(time):
    """
    Function that returns UR script for sleep()

    Args:
        time: float.in s

    Returns:
        script: UR script
    """
    return "sleep({})\n".format(time)


def get_forward_kin(var_name):
    """
    Function that returns UR script for get_forward_kin().

    Transformation from joint space to tool space.

    Args:
        var_name: String. name of variable to store forward kinematics information

    Returns:
        script: UR script
    """
    return "%s = get_forward_kin()\n" % (var_name)


def get_inverse_kin(var_name, ref_plane):
    """
    Function that returns UR script for get_forward_kin().

    Transformation from joint space to tool space.

    Args:
        var_name: String. name of variable to store inverse kinematics information
        ref_plane: Rhino.Geometry Plane. A target plane for calculating pose

    Returns:
        script: UR script
    """
    # TODO: Test

    axis_angle_vector = axis_angle_vector_from_plane_to_plane(ref_plane)

    # Create pose data
    pose = _format_pose(ref_plane, axis_angle_vector)

    # Return UR script
    return "{} = get_inverse_kin({})\n".format(var_name, pose)


def get_joint_positions(var_name):
    """
    Function that returns UR script for get_inverse_kin().
    Transformation from tool space to joint space.

    Args:
        var_name: String. name of variable to store inverse kinematics information
    Returns:
        script: UR script
    """
    # TODO: Test

    return "{} = get_joint_positions()\n".format(var_name)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import Rhino.Geometry as rg


def get_centroid(closed_crv):
    area_obj = rg.AreaMassProperties.Compute(closed_crv)
    return area_obj.Centroid


def reparametrize_crvs(crv_or_crvs, domain=(0, 1)):
    interval = rg.Interval(*domain)
    new_crvs = []

    if not isinstance(crv_or_crvs, list):
        got_list = False
        crv_or_crvs = list(crv_or_crvs)
    else:
        got_list = True

    new_crvs = [crv.DuplicateCurve() for crv in crv_or_crvs]

    for crv in new_crvs:
        crv.Domain = interval

    if not got_list:
        return new_crvs[1]
    else:
        return new_crvs


def array_to_list(array):
    return list(array)


def list_of_arrays_to_list_of_lists(list_of_arrays):
    list_of_lists = []
    for array in list_of_arrays:
        list_of_lists.append(array_to_list(array))
    return list_of_lists



"""
********************************************************************************
compas_rcf.utils
********************************************************************************

.. currentmodule:: compas_rcf.utils

RhinoCommon to COMPAS object conversions
----------------------------------------
Scripts for converting geometry objects between the two frameworks. Most of this
has been integrated into COMPAS v0.15.0 but compas_fab needs earlier versions
of COMPAS.

JSON tools
----------
Utility functions for loading and dumping
:class:`compas_rcf.fabrication.ClayBullet` objects.

Assorted utilities
------------------
When there's no other good place for a bit of code..


"""



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re

from compas.geometry import Frame
from compas.geometry import Point
from compas.geometry import Rotation
from compas.geometry import Transformation
from compas_fab.robots.configuration import Configuration
from compas_fab.robots.ur5 import Robot

from compas_rcf.utils.compas_to_rhino import cgframe_to_rgplane
from compas_rcf.utils.compas_to_rhino import matrix_to_rgtransform

__all__ = ["visualize_urscript"]


def visualize_urscript(script):
    M = [
        [-1000, 0, 0, 0],
        [0, 1000, 0, 0],
        [0, 0, 1000, 0],
        [0, 0, 0, 1],
    ]
    rgT = matrix_to_rgtransform(M)
    cgT = Transformation.from_matrix(M)

    robot = Robot()

    viz_planes = []

    movel_matcher = re.compile(r"^\s*move([lj]).+((-?\d+\.\d+,?\s?){6}).*$")
    for line in script.splitlines():
        mo = re.search(movel_matcher, line)
        if mo:
            if mo.group(1) == "l":  # movel
                ptX, ptY, ptZ, rX, rY, rZ = mo.group(2).split(",")

                pt = Point(float(ptX), float(ptY), float(ptZ))
                pt.transform(cgT)
                frame = Frame(pt, [1, 0, 0], [0, 1, 0])

                R = Rotation.from_axis_angle_vector(
                    [float(rX), float(rY), float(rZ)], pt
                )
                T = matrix_to_rgtransform(R)

                plane = cgframe_to_rgplane(frame)
                plane.Transform(T)

                viz_planes.append(plane)
            else:  # movej
                joint_values = mo.group(2).split(",")
                configuration = Configuration.from_revolute_values(
                    [float(d) for d in joint_values]
                )
                frame = robot.forward_kinematics(configuration)

                plane = cgframe_to_rgplane(frame)
                plane.Transform(rgT)

                viz_planes.append(plane)

    return viz_planes



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from os import name
from os import system
from tkinter import Tk
from tkinter.filedialog import askopenfilename

from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import PygmentsTokens
import pygments
from pygments.lexers.data import YamlLexer

__all__ = ["open_file_dialog", "pygment_yaml", "clear_screen"]

root = Tk()
root.withdraw()


def open_file_dialog(initial_dir="/", file_type=("JSON files", "*.json")):
    filename = askopenfilename(
        initialdir=initial_dir,
        title="Select file",
        filetypes=(file_type, ("all files", "*.*")),
    )
    return filename


def pygment_yaml(yaml):
    lexed_yaml = list(pygments.lex(yaml, lexer=YamlLexer()))
    print_formatted_text(PygmentsTokens(lexed_yaml))


def clear_screen():
    if name == "nt":
        system("cls")
    else:
        system("clear")



"""
********************************************************************************
compas_rcf
********************************************************************************

.. currentmodule:: compas_rcf

python module for MAS DFAB project Rapid Clay Formations

"""
import os

HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))
TEMP = os.path.abspath(os.path.join(HOME, "temp"))

# from https://smarie.github.io/python-getversion/#package-versioning-best-practices
try:
    # -- Distribution mode --
    # import from _version.py generated by setuptools_scm during release
    from ._version import version as __version__
except ImportError:
    try:
        # -- Source mode --
        # use setuptools_scm to get the current version from src using git
        from setuptools_scm import get_version

        __version__ = get_version(HOME)
    except ImportError:
        __version__ = "dev"

__author__ = "Anton T Johansson"
__copyright__ = "MAS DFAB 1920 students and tutors"
__license__ = "MIT License"
__email__ = "anton@tetov.se"

__all__ = ["HOME", "DATA", "DOCS", "TEMP", "__version__"]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math as m

import Rhino.Geometry as rg

from compas_rcf.ur import comm
from compas_rcf.ur import ur_standard
from compas_rcf.ur.visualize_urscript import visualize_urscript
from compas_rcf.utils.util_funcs import list_elem_w_index_wrap

__all__ = ["ur_clay_shooting"]

# UR movement
ROBOT_L_SPEED = 0.6  # m/s
ROBOT_ACCEL = 0.8  # m/s2
ROBOT_SAFE_SPEED = 0.8
ROBOT_J_SPEED = 0.8
BLEND_RADIUS_PUSHING = 0.002  # m

# Tool related variables
TOOL_HEIGHT = 192  # mm
ACTUATOR_IO = 4


def _get_offset_plane(initial_plane, dist, vertical_offset_bool):
    """Generates an offset plane.
    archetypical use: generate entry or exit planes for robotic processes.
    """

    plane_normal = initial_plane.Normal

    if vertical_offset_bool:
        directions = [rg.Vector3d(0, 0, -1), plane_normal]
    else:
        directions = [plane_normal, plane_normal]

    entry_plane = initial_plane.Clone()
    exit_plane = initial_plane.Clone()

    entry_plane.Translate(directions[0] * dist)
    exit_plane.Translate(directions[1] * dist)

    return entry_plane, exit_plane


def _default_movel(plane, blend_radius=0):
    return ur_standard.move_l(
        plane, ROBOT_L_SPEED, ROBOT_ACCEL, blend_radius=blend_radius
    )


def _picking_moves(plane, entry_exit_offset, rotation, vertical_offset_bool):
    script = ""

    entry_plane, exit_plane = _get_offset_plane(
        plane, entry_exit_offset, vertical_offset_bool
    )

    if rotation > 0:
        rotated_plane = plane.Clone()
        rotated_plane.Rotate(m.radians(rotation), plane.Normal)

    script += _default_movel(entry_plane)
    script += _default_movel(plane)
    if rotation > 0:
        script += _default_movel(rotated_plane)
    script += _default_movel(exit_plane)

    return script


def _shooting_moves(
    plane, entry_exit_offset, push_conf, vertical_offset_bool, sleep=0.5
):
    script = ""

    entry_plane, exit_plane = _get_offset_plane(
        plane, entry_exit_offset, vertical_offset_bool
    )

    script += _default_movel(entry_plane)
    script += _default_movel(plane)

    script += ur_standard.set_digital_out(ACTUATOR_IO, True)
    script += ur_standard.sleep(sleep)

    if push_conf["pushing"]:
        script += _push_moves(plane, push_conf, vertical_offset_bool)

    script += _default_movel(exit_plane)

    script += ur_standard.set_digital_out(ACTUATOR_IO, False)

    return script


def _push_moves(plane, push_conf, vertical_offset_bool):

    n = push_conf["n_pushes"]
    dist = push_conf["push_offsets"]
    angle_step = push_conf["angle_steps"]
    rot_axis = push_conf["push_rotation_axis"]

    script = ""

    offset_plane = plane.Clone()
    if vertical_offset_bool:
        direction = rg.Vector3d(0, 0, -1)
    else:
        direction = plane.Normal

    offset_plane.Translate(direction * -dist)

    for i in range(n):
        rot_plane = offset_plane.Clone()

        rot_plane.Rotate(m.radians((i + 1) * angle_step), rot_axis, plane.Origin)

        script += _default_movel(rot_plane, blend_radius=BLEND_RADIUS_PUSHING)

    return script


def _safe_travel_moves(safe_pos_list, reverse=False):

    if reverse:
        safe_pos_list = safe_pos_list[::-1]

    script = ""
    for safe_pos in safe_pos_list:
        script += ur_standard.move_j(safe_pos, ROBOT_SAFE_SPEED, ROBOT_ACCEL)

    return script


def ur_clay_shooting(
    picking_planes,
    placing_planes,
    safe_pos_list,
    dry_run=False,
    push_conf={"pushing": [False]},
    tool_rotation=0,
    picking_rotation=0,
    tool_height_correction=0,
    z_calib_picking=0,
    z_calib_placing=0,
    entry_exit_offset=-40,
    vertical_offset_bool=False,
    viz_planes_bool=False,
    placing_index=0,
):

    reload(comm)  # noqa E0602
    reload(ur_standard)  # noqa E0602

    # extract planes if input is ClayBullets
    if not is_elems_rgPlane(placing_planes):
        placing_planes = [bullet.plane.Clone() for bullet in placing_planes]

    # Create a string object to store script
    script = ""

    # set tcp
    tool_height = TOOL_HEIGHT + tool_height_correction
    script += ur_standard.set_tcp_by_plane_angles(
        0, 0, tool_height, 0.0, 0.0, m.radians(tool_rotation)
    )
    # Ensure actuator is retracted ###
    script += ur_standard.set_digital_out(ACTUATOR_IO, False)

    # Send Robot to an initial known configuration ###
    script += ur_standard.move_j(safe_pos_list[0], ROBOT_ACCEL, ROBOT_SAFE_SPEED)

    # setup instructions

    for key, value in push_conf.iteritems():
        if value is None:
            continue
        if len(value) == len(placing_planes) or len(value) == 1:
            continue

        raise Exception("Mismatched between {} list and placing_plane list".format(key))

    instructions = []
    for i, placing_plane in enumerate(placing_planes):
        instruction = []

        picking_plane = list_elem_w_index_wrap(picking_planes, i)
        instruction.append(picking_plane)

        instruction.append(placing_plane)

        plane_push_conf = {}

        for key, value in push_conf.iteritems():
            if value is not None:
                list_elem = list_elem_w_index_wrap(value, i)
                plane_push_conf.update({key: list_elem})
            else:
                plane_push_conf.update({key: None})

        instruction.append(plane_push_conf)

        instructions.append(instruction)

    # Start general clay fabrication process ###
    for i, instruction in enumerate(instructions):
        picking_plane, placing_plane, push_conf = instruction

        # Pick clay
        if not dry_run:
            # apply z calibration specific to picking station
            picking_plane.Translate(rg.Vector3d(0, 0, z_calib_picking))

            script += _picking_moves(
                picking_plane, entry_exit_offset, picking_rotation, vertical_offset_bool
            )

        # safe moves
        script += _safe_travel_moves(safe_pos_list)

        # apply z calibration specific to placing station
        placing_plane.Translate(rg.Vector3d(0, 0, z_calib_placing))

        script += _shooting_moves(
            placing_plane, entry_exit_offset, push_conf, vertical_offset_bool
        )
        script += ur_standard.UR_log("Bullet {} placed.".format(i + placing_index))

        # safe moves
        script += _safe_travel_moves(safe_pos_list, reverse=True)

    if viz_planes_bool:
        viz_planes = visualize_urscript(script)
    else:
        viz_planes = None

    return comm.concatenate_script(script), viz_planes


def is_elems_rgPlane(obj):
    """ From https://stackoverflow.com/a/18495146
    """
    return bool(obj) and all(isinstance(elem, rg.Plane) for elem in obj)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from subprocess import run
import logging

log = logging.getLogger(__name__)


RUN_KWARGS = {
    "check": True,
    # "capture_output": True,
    # "stdout": subprocess.PIPE,
    # "stderr": subprocess.STDOUT,
    "text": True,
    # "encoding": 'utf-8',
    # "universal_newlines": True
}


def _setup_env_vars(env_vars):
    list_vars = []
    for key in env_vars:
        if os.name == "nt":
            list_vars.append("set")
        list_vars.append("{}={}".format(key.upper(), env_vars[key]))
        list_vars.append("&&")

    return list_vars


def compose_up(
    path, force_recreate=False, remove_orphans=False, ignore_orphans=True, **env_vars
):
    cmd = ["docker-compose", "--file", str(path), "up", "--detach"]

    log.debug("Env vars: {}".format(env_vars))

    if ignore_orphans:
        env_vars.update({"COMPOSE_IGNORE_ORPHANS": "true"})

    if len(env_vars) > 0:
        cmd = _setup_env_vars(env_vars) + cmd
        shell = True
    else:
        shell = False

    if force_recreate:
        cmd.append("--force-recreate")

    if remove_orphans:
        cmd.append("--remove-orphans")

    log.debug("Command to run: {}".format(cmd))

    run(cmd, shell=shell, **RUN_KWARGS)



"""
********************************************************************************
compas_rcf.fabrication
********************************************************************************

.. currentmodule:: compas_rcf.fabrication

Runners & fabrication objects
-----------------------------
This modules contains runners; scripts for running a fabrication process using a
digital fabrication method. It also contains fabrication objects, most significantly
:class:`compas_rcf.fabrication.ClayBullet` which is our data representation of our
discrete building blocks.

"""



"""Not ready yet
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import socket
import time
from pathlib import Path

from compas_rcf.ur.helpers import turn_off_outputs_program
from compas_rcf.ur.urscript_wrapper import movel
from compas_rcf.ur.urscript_wrapper import set_DO
from compas_rcf.ur.urscript_wrapper import set_TCP
from compas_rcf.ur.urscript_wrapper import socket_close
from compas_rcf.ur.urscript_wrapper import socket_open
from compas_rcf.ur.urscript_wrapper import socket_send_string
from compas_rcf.ur.urscript_wrapper import textmsg
from compas_rcf.utils.database import get_bullets

# ===============================================================

# GLOBALS
# ===============================================================

DEBUG = False

FILE_DIR = Path(__file__).parent.absolute()
REPO_PATH = Path(__file__).parents[2].absolute()
FILE_NAME = "commands_group_04.json"
SERVER_ADDRESS = "192.168.10.11"
SERVER_PORT = 30003
UR_IP = "192.168.10.10"
UR_SERVER_PORT = 30002

ROBOT_L_SPEED = 0.6  # m/s
ROBOT_ACCEL = 0.8  # m/s2
ROBOT_SAFE_SPEED = 0.8
ROBOT_J_SPEED = 0.8
ZONING = 0.0  # m

AIR_PRESSURE_DO = 0

UR_PROGRAM_FOOTER = ""
UR_PROGRAM_FOOTER += socket_open(SERVER_ADDRESS, SERVER_PORT)
UR_PROGRAM_FOOTER += textmsg("End program")
UR_PROGRAM_FOOTER += socket_send_string("Script finished")
UR_PROGRAM_FOOTER += socket_close()
UR_PROGRAM_FOOTER += "end\n"
UR_PROGRAM_FOOTER += "program()\n"

# ===============================================================

# UTIL
# ===============================================================

LOG_FILE = Path.joinpath(FILE_DIR, "main_direct_send_group_04.log")
UR_SCRIPT_LOG = Path.joinpath(FILE_DIR, "main_direct_send_group_04.ur_log.log")
logging.basicConfig(
    filename=LOG_FILE,
    filemode="a",
    level=logging.DEBUG,
    format="%(asctime)s:%(levelname)s:%(funcName)s:%(message)s",
)


def start_log() -> None:
    logging.debug("Start script")
    logging.debug("GLOBALS")
    logging.debug("FILE_DIR: {}".format(FILE_DIR))
    logging.debug("REPO_PATH: {}".format(REPO_PATH))
    logging.debug("FILE_NAME: {}".format(FILE_NAME))
    logging.debug("SERVER_ADDRESS: {}".format(SERVER_ADDRESS))
    logging.debug("SERVER_PORT: {}".format(SERVER_PORT))
    logging.debug("UR_IP: {}".format(UR_IP))
    logging.debug("UR_SERVER_PORT: {}".format(UR_SERVER_PORT))
    logging.debug("AIR_PRESSURE_DO: {}".format(AIR_PRESSURE_DO))


# ===============================================================

# PARSE COMMANDS FROM JSON
# ===============================================================


"""
def parse_json(file_name):
    file = Path.joinpath(FILE_DIR.parent, file_name)

    with file.open(mode='r') as f:
        data = json.load(f)

    start_at_safe_pt = data['start_at_safe_pt']
    len_command = data['len_command']
    gh_commands = data['gh_commands']

    commands = format_commands(gh_commands, len_command)

    # validate_commands(commands)

    print("We have %d commands to send" % len(commands))

    return commands, start_at_safe_pt
"""


# ===============================================================

# UR SCRIPT COMMANDS
# ===============================================================

# ===============================================================

# ===============================================================

# UR SCRIPT PROGRAMS
# ===============================================================


def generate_bullet_placing_program(bullet):
    script = []
    script += "def program():\n"
    script += set_TCP(bullet.tcp)

    for i, frame in enumerate(bullet.pre_frames):
        script += movel(frame, accel=ROBOT_ACCEL, speed=ROBOT_L_SPEED, radius=ZONING)

    script += movel(
        bullet.placement_frame, accel=ROBOT_ACCEL, speed=ROBOT_L_SPEED, radius=ZONING
    )
    script += set_DO(AIR_PRESSURE_DO, True)
    script += textmsg("Air pressure on")

    for i, frame in enumerate(bullet.post_frames):
        script += movel(frame, accel=ROBOT_ACCEL, speed=ROBOT_L_SPEED, radius=ZONING)

    script += UR_PROGRAM_FOOTER

    if DEBUG:
        with open(UR_SCRIPT_LOG, mode="w") as f:
            f.write(script.decode("utf-8"))

    return script


# ===============================================================

# ===============================================================


def main() -> None:
    start_log()

    bullets = get_bullets()

    send_socket = socket.create_connection((UR_IP, UR_SERVER_PORT), timeout=2)
    send_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    logging.debug("Sockets setup")

    # make server
    recv_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    recv_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    # Bind the socket to the port
    recv_socket.bind((SERVER_ADDRESS, SERVER_PORT))

    for bullet in bullets:
        script = generate_bullet_placing_program(bullet)

        print("Sending placing commands for bullet id {}".format(bullet.id))

        # send file
        script = script.encode()
        send_socket.send(script)
        logging.debug("File sent")

        # Listen for incoming connections
        recv_socket.listen(1)
        while True:
            logging.debug("Waiting for accept")
            connection, client_address = recv_socket.accept()
            logging.debug("Received accept from: {}".format(client_address))
            break

    recv_socket.close()

    script = turn_off_outputs_program()

    send_socket.send(script)

    time.sleep(5)

    send_socket.close()
    print("Program done ...")


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import confuse

from compas_rcf.abb.helpers import zone_dict

__all__ = ["ZoneDataTemplate", "abb_rcf_conf_template", "fabrication_conf"]


class ZoneDataTemplate(confuse.Template):
    def __init__(self, default=confuse.REQUIRED):
        super(ZoneDataTemplate, self).__init__(default=default)

    def convert(self, value, view):
        if isinstance(value, (int, float)):
            if not -1 >= value >= 2000:  # arbitrary max value
                self.fail("ZoneData needs to be from -1 to 2000", view)
            return value
        if value.upper() not in zone_dict.keys():
            self.fail(
                "ZoneData must match one of {0}".format(", ".join(zone_dict.keys())),
                view,
            )
        return zone_dict[value.upper()]


abb_rcf_conf_template = {
    # Two following is set by command line arguments
    "debug": confuse.TypeTemplate(bool, default=False),
    "verbose": confuse.TypeTemplate(bool, default=False),
    # is_target_real is set either by command line argument, during run or in conf file
    "target": confuse.TypeTemplate(str, default=None),
    "wobjs": {"picking_wobj_name": str, "placing_wobj_name": str},
    "tool": {
        "tool_name": str,
        "io_needles_pin": str,
        "grip_state": int,
        "release_state": int,
        "wait_before_io": confuse.Number(default=2),
        "wait_after_io": confuse.Number(default=0.5),
    },
    "speed_values": {
        "speed_override": confuse.Number(default=100),
        "speed_max_tcp": float,
        "accel": float,
        "accel_ramp": confuse.Number(default=100),
    },
    "safe_joint_positions": {
        "start": confuse.Sequence([float] * 6),
        "end": confuse.Sequence([float] * 6),
    },
    "movement": {
        "offset_distance": float,
        "speed_placing": float,
        "speed_picking": float,
        "speed_travel": float,
        "zone_travel": ZoneDataTemplate(),
        "zone_pick": ZoneDataTemplate(),
        "zone_place": ZoneDataTemplate(),
    },
    "pick": {
        "origin_grid": {
            "x": confuse.Number(default=100),
            "y": confuse.Number(default=100),
        },
        "xnum": int,
        "ynum": int,
        "grid_spacing": float,
        "compression_height_factor": confuse.Number(default=0.95),
        "xaxis": confuse.Sequence([float] * 3),
        "yaxis": confuse.Sequence([float] * 3),
    },
}

fabrication_conf = confuse.LazyConfig("FabricationRunner", __name__)



"""Most of these are in compas >=0.15 but compas_fab is not there yet
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import compas.geometry as cg
from compas import IPY

if IPY:
    import Rhino.Geometry as rg


def cgpoint_to_rgpoint(pt):  # type: (cg.Point) -> rg.Point3d
    """Convenience function to convert a compas.geometry.Point object to the
       corresponding COMPAS object

    Parameters
    ----------
    compas.geometry.Point
        Point object to convert

    Returns
    -------
    Rhino.Geometry.Point3d
        Resulting Point3d object
    """
    return rg.Point3d(*pt.data)


def cgvector_to_rgvector(v):  # type: (cg.Vector) -> rg.Vector3d
    """Convenience function to convert a compas.geometry.Vector object to the
       corresponding COMPAS object

    Parameters
    ----------
    compas.geometry.Vector
        Vector object to convert

    Returns
    -------
    Rhino.Geometry.Vector3d
        Resulting Vector3d object
    """
    # TODO: Implement in compas
    return rg.Vector3d(*v.data)


def cgline_to_rgline(line):  # type: (cg.Line) -> rg.Line
    """Convenience function to convert a compas.geometry.Line object to the
       corresponding COMPAS object

    Parameters
    ----------
    compas.geometry.Line
        Point object to convert

    Returns
    -------
    Rhino.Geometry.Line
        Resulting Line object
    """
    return rg.Line(cgpoint_to_rgpoint(line.start), cgpoint_to_rgpoint(line.end))


def cgplane_to_rgplane(cgplane):  # type: (cg.Plane) -> rg.Plane
    """Convenience function to convert a compas.geometry.Plane object to the
       corresponding Rhino.Geometry object
    Parameters
    ----------
    compas.geometry.Plane
        Plane to convert
    Returns
    -------
    Rhino.Geometry.Plane
        Resulting plane
    """
    return rg.Plane(
        cgpoint_to_rgpoint(cgplane.point), cgvector_to_rgvector(cgplane.normal)
    )


def cgframe_to_rgplane(frame):  # type: (cg.Frame) -> rg.Plane
    """Convenience function to convert a compas.geometry.Frame object to a
       Rhino.Geometry.Plane object
    Parameters
    ----------
    compas.geometry.Frame
        Frame to convert
    Returns
    -------
    Rhino.Geometry.Plane
        Resulting plane
    """
    plane = cg.Plane(frame.point, frame.normal)
    return cgplane_to_rgplane(plane)


def matrix_to_rgtransform(M):
    rgM = rg.Transform()
    for i, row in enumerate(M):
        for j, val in enumerate(row):
            rgM[i, j] = val
    return rgM


if __name__ == "__main__":

    # test cgpoint_to_rgpoint
    point = cgpoint_to_rgpoint(cg.Point(1, 2, 3))
    assert point.Z == 3.0

    # test cgvector_to_rgvector
    vector = cgvector_to_rgvector(cg.Vector(5, 1, 9))
    assert vector.Unitize()

    # test cgline_to_rgline
    line = cgline_to_rgline(cg.Line([1, 2, 3], [3, 2, 1]))
    assert line.Direction == rg.Vector3d(2, 0, -2)

    # test frame_to_plane
    plane = cgframe_to_rgplane(cg.Frame([1, 3, -1], [1, 1, 2], [0, 1, 1]))
    assert isinstance(plane.Normal, rg.Vector3d)

    # matrix_to_rgtransform
    R = cg.Rotation.from_basis_vectors([1, 2, 0], [2, 1, 3])
    assert isinstance(matrix_to_rgtransform(R), rg.Transform)



"""Most of these are in compas >=0.15 but compas_fab is not there yet."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import compas.geometry as cg
import Rhino.Geometry as rg


def rgpoint_to_cgpoint(pt):  # type: (rg.Point3d) -> cg.Point
    """Convenience function to convert Rhino.Geometry.Point3d object to the
       corresponding COMPAS object

    Parameters
    ----------
    Rhino.Geometry.Point3d
        Plane object to convert

    Returns
    -------
    compas.geometry.Point
        Resulting point object
    """
    return cg.Point(pt.X, pt.Y, pt.Z)


def rgvector_to_cgvector(v):  # type: (rg.Point3d) -> cg.Vector
    """Convenience function to convert Rhino.Geometry.Vector3d object to the
       corresponding COMPAS object

    Parameters
    ----------
    Rhino.Geometry.Vector3d
        Vector object to convert

    Returns
    -------
    compas.geometry.Vector
        Resulting vector object
    """
    return cg.Vector(v.X, v.Y, v.Z)


def rgline_to_cgline(line):  # type: (rg.Line) -> cg.Line
    """Convenience function to convert Rhino.Geometry.Line object to the
       corresponding COMPAS object

    Parameters
    ----------
    Rhino.Geometry.Line
        Line object to convert

    Returns
    -------
    compas.geometry.Line
        Resulting line object
    """
    return cg.Line(rgpoint_to_cgpoint(line.From), rgpoint_to_cgpoint(line.To))


def rgplane_to_cgplane(plane):  # type: (rg.Plane) -> cg.Plane
    """Convenience function to convert Rhino.Geometry.Plane object to the
       corresponding compas.geometry object
    Parameters
    ----------
    Rhino.Geometry.Plane
        Plane object to convert
    Returns
    -------
    compas.geometry.Plane
        Resulting plane object

    Notes
    -----
    Unlike a Rhino.Geometry plane the compas.geometry plane does not store xaxis
    and yaxis vector. See :meth:`compas.geometry.Frame.from_plane` docstring.
    """
    return cg.Plane(
        rgpoint_to_cgpoint(plane.Origin), rgvector_to_cgvector(plane.Normal)
    )


def rgplane_to_cgframe(plane):  # type: (rg.Plane) -> cg.Frame
    """Convenience function to convert Rhino.Geometry.Plane object to a
       compas.geometry.Frame object
    Parameters
    ----------
    Rhino.Geometry.Plane
        Plane object to convert
    Returns
    -------
    compas.geometry.Frame
        Resulting frame object
    """
    pt = rgpoint_to_cgpoint(plane.Origin)
    xaxis = rgvector_to_cgvector(plane.XAxis)
    yaxis = rgvector_to_cgvector(plane.YAxis)
    return cg.Frame(pt, xaxis, yaxis)


def rgtransform_to_matrix(rgM):
    M = [[rgM.Item[i, j] for j in range(4)] for i in range(4)]
    return M


if __name__ == "__main__":

    # rgpoint_to_cgpoint
    pt = rgpoint_to_cgpoint(rg.Point3d(3, 2, 1))
    assert pt.data == [3.0, 2.0, 1.0]

    # rgvector_to_cgvector
    v = rgvector_to_cgvector(rg.Vector3d(3, 2, 1))
    assert v.length > 3.7 and v.length < 3.8

    # rgline_cgline
    line = rgline_to_cgline(rg.Line(rg.Point3d(3, 2, 1), rg.Vector3d(1, 1, 0), 5.0))
    assert isinstance(line.midpoint.z, float)

    # rgplane_to_cgframe
    frame = rgplane_to_cgframe(rg.Plane(rg.Point3d(1, 3, 2), rg.Vector3d(2, -1, 1)))
    frame.quaternion.__repr__ == "Quaternion(0.713799, 0.462707, 0.285969, 0.441152)"

    # rgtransform_to_matrix
    matrix = rg.Transform.ZeroTransformation
    assert rgtransform_to_matrix(matrix) == [
        [0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 1.0],
    ]



# -*- coding: utf-8 -*-
from __future__ import print_function

import contextlib
import glob
import os
import sys
from shutil import rmtree

from invoke import task

try:
    input = raw_input
except NameError:
    pass

BASE_FOLDER = os.path.dirname(__file__)


class Log(object):
    def __init__(self, out=sys.stdout, err=sys.stderr):
        self.out = out
        self.err = err

    def flush(self):
        self.out.flush()
        self.err.flush()

    def write(self, message):
        self.flush()
        self.out.write(message + "\n")
        self.out.flush()

    def info(self, message):
        self.write("[INFO] %s" % message)

    def warn(self, message):
        self.write("[WARN] %s" % message)


log = Log()


def confirm(question):
    while True:
        response = input(question).lower().strip()

        if not response or response in ("n", "no"):
            return False

        if response in ("y", "yes"):
            return True

        print("Focus, kid! It is either (y)es or (n)o", file=sys.stderr)


@task(default=True)
def help(ctx):
    """Lists available tasks and usage."""
    ctx.run("invoke --list")
    log.write('Use "invoke -h <taskname>" to get detailed help for a task.')


@task(
    help={
        "docs": "True to clean up generated documentation, otherwise False",
        "bytecode": "True to clean up compiled python files, otherwise False.",
        "builds": "True to clean up build/packaging artifacts, otherwise False.",
    }
)
def clean(ctx, docs=True, bytecode=True, builds=True):
    """Cleans the local copy from compiled artifacts."""

    with chdir(BASE_FOLDER):
        if builds:
            ctx.run("python setup.py clean")

        if bytecode:
            for root, dirs, files in os.walk(BASE_FOLDER):
                for f in files:
                    if f.endswith(".pyc"):
                        os.remove(os.path.join(root, f))
                if ".git" in dirs:
                    dirs.remove(".git")

        folders = []

        if docs:
            folders.append("docs/api/generated")

        folders.append("dist/")

        if bytecode:
            for t in ("src", "tests"):
                folders.extend(glob.glob("{}/**/__pycache__".format(t), recursive=True))

        if builds:
            folders.append("build/")
            folders.append("src/compas_rcf.egg-info/")

        for folder in folders:
            rmtree(os.path.join(BASE_FOLDER, folder), ignore_errors=True)


@task(
    help={
        "rebuild": "True to clean all previously built docs before starting.",
        "doctest": "True to run doctests.",
        "check_links": "True to check all web links in docs for validity.",
    }
)
def docs(ctx, doctest=False, rebuild=True, check_links=False):
    """Builds package's HTML documentation."""

    if rebuild:
        clean(ctx)

    with chdir(BASE_FOLDER):
        if doctest:
            ctx.run("sphinx-build -E -b doctest docs build/docs")

        ctx.run("sphinx-build -E -b html docs build/docs")

        if check_links:
            ctx.run("sphinx-build -E -b linkcheck docs build/docs")


@task()
def check(ctx):
    """Check the consistency of documentation, coding style and a few other things."""

    with chdir(BASE_FOLDER):
        log.write("Checking metadata...")
        ctx.run("python setup.py check --strict --metadata")

        log.write("Running flake8 python linter...")
        ctx.run("flake8 --count --statistics src tests")

        log.write("Checking python imports...")
        ctx.run("isort --check-only --diff --recursive src tests setup.py")


@task(help={"checks": "True to run all checks before testing, otherwise False."})
def test(ctx, checks=False, doctest=False):
    """Run all tests."""
    if checks:
        check(ctx)

    with chdir(BASE_FOLDER):
        cmd = ["pytest"]
        if doctest:
            cmd.append("--doctest-modules")

        ctx.run(" ".join(cmd))


@task
def prepare_changelog(ctx):
    """Prepare changelog for next release."""
    UNRELEASED_CHANGELOG_TEMPLATE = (
        "## Unreleased\n\n### Added\n\n### Changed\n\n### Removed\n\n\n## "
    )

    with chdir(BASE_FOLDER):
        # Preparing changelog for next release
        with open("CHANGELOG.md", "r+") as changelog:
            content = changelog.read()
            changelog.seek(0)
            changelog.write(content.replace("## ", UNRELEASED_CHANGELOG_TEMPLATE, 1))

        ctx.run(
            'git add CHANGELOG.md && git commit -m "Prepare changelog for next release"'
        )


@task
def build(ctx):
    """Build project."""
    ctx.run("python setup.py clean --all sdist bdist_wheel")


@task(help={"new_version": "Version number to release"})
def release(ctx, new_version):
    """Releases the project in one swift command!"""
    # Run checks
    ctx.run("invoke check test build docs")

    # TODO: check if dirty

    # Bump version and git tag it
    ctx.run("git -s tag {}".format(new_version))

    prepare_changelog(ctx)


@contextlib.contextmanager
def chdir(dirname=None):
    current_dir = os.getcwd()
    try:
        if dirname is not None:
            os.chdir(dirname)
        yield
    finally:
        os.chdir(current_dir)



"""Helpers for scripts interacting with ABB robots or RAPID code."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_rrc import Noop
from compas_rrc import FeedbackLevel

from compas_rcf import HERE

# Describes the valid zone data definitions.
zone_dict = {
    "FINE": -1,
    "Z0": 0,
    "Z1": 1,
    "Z5": 5,
    "Z10": 10,
    "Z15": 15,
    "Z20": 20,
    "Z30": 30,
    "Z40": 40,
    "Z50": 50,
    "Z60": 60,
    "Z80": 80,
    "Z100": 100,
    "Z150": 150,
    "Z200": 200,
}

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path

pkg_dir = Path(HERE)

_compose_folder = pkg_dir / "utils" / "docker" / "compose_files" / "abb"
docker_compose_paths = {
    "base": _compose_folder / "base-docker-compose.yml",
    "abb_driver": _compose_folder / "abb-driver-docker-compose.yml",
}
robot_ips = {"real": "192.168.125.1", "virtual": "host.docker.internal"}


def ping(client, timeout=10):
    feedback = client.send(Noop(feedback_level=FeedbackLevel.DONE))

    try:
        return feedback.result(timeout=timeout)
    except Exception as e:
        if e.args[0] == "Timeout: future result not available":
            raise TimeoutError(e.args)
        else:
            raise



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json

from compas_rcf.fabrication.clay_obj import ClayBullet

all = ["load_bullets"]


def load_bullets(file_path):

    with open(file_path, mode="r") as fp:
        json_string = json.load(fp)

    clay_bullets = []
    for dict_ in json_string:
        clay_bullets.append(ClayBullet.from_data(dict_))

    return clay_bullets



"""Fabrication runner for Rapid Clay Fabrication project for fullscale structure.

Run from command line using :code:`python -m compas_rcf.fabrication.abb_rcf_runner`
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import sys
import time
from datetime import datetime
from pathlib import Path
import json

from colorama import Fore
from colorama import Style
from colorama import init
from compas.geometry import Frame
from compas.geometry import Point
from compas.geometry import Vector
from compas_fab.backends.ros import RosClient
from compas_rrc import AbbClient
from compas_rrc import CustomInstruction
from compas_rrc import MoveToFrame
from compas_rrc import MoveToJoints
from compas_rrc import PrintText
from compas_rrc import SetAcceleration
from compas_rrc import SetDigital
from compas_rrc import SetMaxSpeed
from compas_rrc import SetTool
from compas_rrc import SetWorkObject
from compas_rrc import WaitTime
from compas_rrc import StartWatch
from compas_rrc import StopWatch
from compas_rrc import ReadWatch

from compas_rcf import __version__
from compas_rcf.abb.helpers import docker_compose_paths
from compas_rcf.abb.helpers import ping
from compas_rcf.abb.helpers import robot_ips
from compas_rcf.fabrication.conf import abb_rcf_conf_template
from compas_rcf.fabrication.conf import fabrication_conf
from compas_rcf.utils import ui
from compas_rcf.utils.docker import compose_up
from compas_rcf.utils.json_ import load_bullets
from compas_rcf.utils.util_funcs import get_offset_frame
from compas_rcf.fabrication.clay_obj import ClayBulletEncoder

if sys.version_info[0] < 2:
    raise Exception("This module requires Python 3")
else:
    import questionary

################################################################################
# Globals                                                                      #
################################################################################

ROBOT_CONTROL_FOLDER_DRIVE = Path(
    "G:\\Shared drives\\2020_MAS\\T2_P1\\02_Groups\\Phase2\\rcf_fabrication\\02_robot_control"  # noqa E501
)
DEFAULT_CONF_DIR = ROBOT_CONTROL_FOLDER_DRIVE / "05_fabrication_confs"
DEFAULT_JSON_DIR = ROBOT_CONTROL_FOLDER_DRIVE / "04_fabrication_data_jsons"
DEFAULT_LOG_DIR = ROBOT_CONTROL_FOLDER_DRIVE / "06_fabrication_logs"

# Define external axis, will not be used but required in move cmds
EXTERNAL_AXIS_DUMMY: list = []

################################################################################
# Programs                                                                     #
################################################################################


def initial_setup(client):
    """Pre fabrication setup, speed, acceleration, tool, work object and initial pose.

    Parameters
    ----------
    client : :class:`compas_rrc.AbbClient`
    """
    send_grip_release(client, CONF.tool.release_state)

    client.send(SetTool(CONF.tool.tool_name))
    logging.debug("Tool {} set.".format(CONF.tool.tool_name))
    client.send(SetWorkObject(CONF.wobjs.placing_wobj_name))
    logging.debug("Work object {} set.".format(CONF.wobjs.placing_wobj_name))

    # Set Acceleration
    client.send(SetAcceleration(CONF.speed_values.accel, CONF.speed_values.accel_ramp))
    logging.debug("Acceleration values set.")

    # Set Max Speed
    client.send(
        SetMaxSpeed(CONF.speed_values.speed_override, CONF.speed_values.speed_max_tcp)
    )
    logging.debug("Speed set.")

    # Initial configuration
    client.send(
        MoveToJoints(
            CONF.safe_joint_positions.start,
            EXTERNAL_AXIS_DUMMY,
            CONF.movement.speed_travel,
            CONF.movement.zone_travel,
        )
    )
    logging.debug("Sent move to safe joint position")


def shutdown_procedure(client):
    """Post fabrication procedure, end pose and closing and termination of client.

    Parameters
    ----------
    client : :class:`compas_rrc.AbbClient`
    """
    send_grip_release(client, CONF.tool.release_state)

    client.send(
        MoveToJoints(
            CONF.safe_joint_positions.end,
            EXTERNAL_AXIS_DUMMY,
            CONF.movement.speed_travel,
            CONF.movement.zone_travel,
        )
    )

    client.send_and_wait(PrintText("Finished"))

    # Close client
    client.close()
    client.terminate()


def send_picking(client, picking_frame):
    """Send movement and IO instructions to pick up a clay bullet.

    Parameters
    ----------
    client : :class:`compas_rrc.AbbClient`
    picking_frame : compas.geometry.Frame
        Target frame to pick up bullet
    """
    if CONF.target == "virtual":
        # Custom instruction create a clay bullet in RobotStudio
        # TODO: Create bullet at picking point
        client.send(CustomInstruction("r_A057_RS_Create_Bullet"))

    # change work object before picking
    client.send(SetWorkObject(CONF.wobjs.picking_wobj_name))

    # pick bullet
    offset_picking = get_offset_frame(picking_frame, CONF.movement.offset_distance)

    # start watch
    client.send(StartWatch())

    client.send(
        MoveToFrame(
            offset_picking, CONF.movement.speed_travel, CONF.movement.zone_travel
        )
    )

    client.send(
        MoveToFrame(
            picking_frame, CONF.movement.speed_travel, CONF.movement.zone_travel
        )
    )

    send_grip_release(client, CONF.tool.grip_state)

    client.send(
        MoveToFrame(
            offset_picking, CONF.movement.speed_picking, CONF.movement.zone_pick
        )
    )

    client.send(StopWatch())
    return client.send(ReadWatch())


def send_placing(client, bullet):
    """Send movement and IO instructions to place a clay bullet.

    Parameters
    ----------
    client : :class:`compas_rrc.AbbClient`
    picking_frame : compas.geometry.Frame
        Target frame to pick up bullet
    """
    logging.debug("Location frame: {}".format(bullet.location))

    # change work object before placing
    client.send(SetWorkObject(CONF.wobjs.placing_wobj_name))

    # add offset placing plane to pre and post frames

    top_bullet_frame = get_offset_frame(bullet.location, bullet.height)
    offset_placement = get_offset_frame(top_bullet_frame, CONF.movement.offset_distance)

    # start watch
    client.send(StartWatch())

    # Safe pos then vertical offset
    for frame in bullet.trajectory_to:
        client.send(
            MoveToFrame(frame, CONF.movement.speed_travel, CONF.movement.zone_travel)
        )

    client.send(
        MoveToFrame(
            offset_placement, CONF.movement.speed_travel, CONF.movement.zone_travel
        )
    )
    client.send(
        MoveToFrame(
            top_bullet_frame, CONF.movement.speed_placing, CONF.movement.zone_place
        )
    )

    send_grip_release(client, CONF.tool.release_state)

    client.send(
        MoveToFrame(
            bullet.placement_frame,
            CONF.movement.speed_placing,
            CONF.movement.zone_place,
        )
    )

    client.send(
        MoveToFrame(
            offset_placement, CONF.movement.speed_travel, CONF.movement.zone_travel
        )
    )

    # offset placement frame then safety frame
    for frame in bullet.trajectory_from:
        client.send(
            MoveToFrame(frame, CONF.movement.speed_travel, CONF.movement.zone_travel)
        )

    client.send(StopWatch())
    return client.send(ReadWatch())


def send_grip_release(client, do_state):
    """Grip or release using RCF tool, either in simulation or on real robot.

    If script target is real robot this will set the digital output on the robot,
    and if the target is virtual robot it will use RobotStudio code to visualize
    clay bullet gripping and releasing

    Parameters
    ----------
    client : :class:`compas_rrc.AbbClient`
    do_state : int (0 or 1)
        Value to set DO to
    """
    if CONF.target == "real":
        client.send(WaitTime(CONF.tool.wait_before_io))
        client.send(SetDigital(CONF.tool.io_needles_pin, do_state))
        client.send(WaitTime(CONF.tool.wait_after_io))
    else:
        # Custom instruction can grip a bullet in RobotStudio
        # note the tool tip must touch the bullet
        if do_state == CONF.tool.grip_state:
            client.send(CustomInstruction("r_A057_RS_ToolGrip"))
        else:
            client.send(CustomInstruction("r_A057_RS_ToolRelease"))

    logging.debug(
        "Signal sent to {}".format(
            "grip" if do_state == CONF.tool.grip_state else "release"
        )
    )


################################################################################
# Non programs
################################################################################


def get_settings():
    """Print and prompts user for changes to default configuration.

    Parameters
    ----------
    target_select : str ('real' or 'virtual')
        Target for script, either virtual robot controller or real. From argparse.
    """
    init(autoreset=True)

    load_or_default = questionary.select(
        "Load config or use default?", choices=["Default", "Load"], default="Default"
    ).ask()

    if load_or_default == "Load":
        conf_file = ui.open_file_dialog(
            initial_dir=DEFAULT_CONF_DIR, file_type=("YAML files", "*.yaml")
        )
        fabrication_conf.set_file(conf_file)
        logging.info("Configuration loaded from {}".format(conf_file))
    else:
        fabrication_conf.read(defaults=True, user=False)
        logging.info("Default configuration loaded from package")

    if not fabrication_conf["target"].exists():
        question = questionary.select(
            "Target?", choices=["Virtual robot", "Real robot"], default="Virtual robot"
        ).ask()
        print(question)
        fabrication_conf["target"] = "real" if question == "Real robot" else "virtual"
        print(fabrication_conf["target"])

    logging.info(
        "Target is {} controller.".format(fabrication_conf["target"].get().upper())
    )

    # At this point the conf is considered set, if changes needs to happen after
    # this point CONF needs to be set again. There's probably a better way though.
    global CONF
    CONF = fabrication_conf.get(abb_rcf_conf_template)

    print(Fore.CYAN + Style.BRIGHT + "Configuration")

    ui.pygment_yaml(fabrication_conf.dump())

    conf_ok = questionary.confirm("Configuration correct?").ask()
    if not conf_ok:
        logging.critical("Program exited because user didn't confirm config")
        print("Exiting.")
        sys.exit()


def pick_frame_from_grid(index, bullet_height):
    """Get next picking frame.

    Parameters
    ----------
    index : int
        Counter to iterate through picking positions.
    bullet_height : float
        Height of bullet to pick up.

    Returns
    -------
    list of `class`:compas.geometry.Frame
    """
    # If index is larger than amount on picking plate, start from zero again
    index = index % (CONF.pick.xnum * CONF.pick.ynum)

    xpos = index % CONF.pick.xnum
    ypos = index // CONF.pick.xnum

    x = CONF.pick.origin_grid.x + xpos * CONF.pick.grid_spacing
    y = CONF.pick.origin_grid.y + ypos * CONF.pick.grid_spacing
    z = bullet_height * CONF.pick.compression_height_factor

    frame = Frame(Point(x, y, z), Vector(*CONF.pick.xaxis), Vector(*CONF.pick.yaxis))
    logging.debug("Picking frame {:03d}: {}".format(index, frame))
    return frame


################################################################################
# Script runner                                                                #
################################################################################
def abb_run(cmd_line_args):
    """Fabrication runner, sets conf, reads json input and runs fabrication process."""
    print("\ncompas_rfc abb runner\n")

    fabrication_conf.set_args(cmd_line_args)

    ############################################################################
    # Logging setup                                                            #
    ############################################################################
    timestamp_file = datetime.now().strftime("%Y%m%d-%H.%M_rcf_abb.log")
    log_file = DEFAULT_LOG_DIR / timestamp_file

    handlers = []

    if not fabrication_conf["skip_logfile"]:
        handlers.append(logging.FileHandler(log_file, mode="a"))

    if fabrication_conf["verbose"]:
        handlers.append(logging.StreamHandler(sys.stdout))

    logging.basicConfig(
        level=logging.DEBUG if fabrication_conf["debug"] else logging.INFO,
        format="%(asctime)s:%(levelname)s:%(funcName)s:%(message)s",
        handlers=handlers,
    )

    logging.info("compas_rcf version: {}".format(__version__))
    logging.debug("argparse input: {}".format(cmd_line_args))
    logging.debug("config after set_args: {}".format(fabrication_conf))

    ############################################################################
    # CONF setup                                                              #
    ############################################################################
    get_settings()
    logging.info("Fabrication configuration:\n{}".format(fabrication_conf.dump()))

    ############################################################################
    # Docker setup                                                            #
    ############################################################################
    compose_up(docker_compose_paths["base"], remove_orphans=False)
    logging.debug("Compose up base")
    ip = robot_ips[CONF.target]
    compose_up(docker_compose_paths["abb_driver"], ROBOT_IP=ip)
    logging.debug("Compose up abb_driver")

    ############################################################################
    # Load fabrication data                                                    #
    ############################################################################
    json_path = Path(ui.open_file_dialog(initial_dir=DEFAULT_JSON_DIR))

    clay_bullets = load_bullets(json_path)
    logging.info("Fabrication data read from: {}".format(json_path))
    logging.info("{} items in clay_bullets.".format(len(clay_bullets)))

    ############################################################################
    # Create Ros Client                                                        #
    ############################################################################
    ros = RosClient()

    ############################################################################
    # Create ABB Client                                                        #
    ############################################################################
    abb = AbbClient(ros)
    abb.run()
    logging.debug("Connected to ROS")

    ############################################################################
    # Connection check                                                         #
    ############################################################################
    ip = robot_ips[CONF.target]
    for i in range(3):
        try:
            logging.debug("Pinging robot")
            ping(abb, timeout=5)
            logging.debug("Breaking loop after successful ping")
            break
        except TimeoutError:
            logging.info("No response from controller, restarting abb-driver service.")
            compose_up(
                docker_compose_paths["abb_driver"], force_recreate=True, ROBOT_IP=ip
            )
            logging.debug("Compose up for abb_driver with robot-ip={}".format(ip))
            time.sleep(10)
    else:
        raise TimeoutError("Failed to connect to robot")

    ############################################################################
    # Set speed, accel, tool, wobj and move to start pos                       #
    ############################################################################
    initial_setup(abb)

    ############################################################################
    # Fabrication loop                                                         #
    ############################################################################

    for i, bullet in enumerate(clay_bullets):
        logging.info(
            "Bullet {:03d}/{:03d} with id {}".format(i, len(clay_bullets), bullet.id)
        )

        pick_frame = pick_frame_from_grid(i, bullet.height)

        # Pick bullet
        pick_future = send_picking(abb, pick_frame)

        # Place bullet
        place_future = send_placing(abb, bullet)

        cycle_time = pick_future.result() + place_future.result()

        bullet.cycle_time = cycle_time
        logging.debug("Cycle time was {}".format(bullet.cycle_time))
        bullet.placed = time.time()
        logging.debug("Time placed was {}".format(bullet.placed))

    ############################################################################
    # Shutdown procedure                                                       #
    ############################################################################

    done_json = DEFAULT_JSON_DIR / "00_done" / json_path.name

    with done_json.open(mode="w") as fp:
        json.dump(clay_bullets, fp, cls=ClayBulletEncoder)

    logging.debug("Saved placed bullets.")

    logging.info("Finished program with {} bullets.".format(len(clay_bullets)))
    shutdown_procedure(abb)


if __name__ == "__main__":
    """Entry point and argument handling
    """
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-t",
        "--target",
        choices=["real", "virtual"],
        help="Set fabrication runner target.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Prints logging messages to console.",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Add DEBUG level messages to logfile, and print them on console if --verbose is set.",  # noqa E501
    )
    parser.add_argument(
        "--skip-logfile", action="store_true", help="Don't send log messages to file.",
    )

    args = parser.parse_args()

    abb_run(args)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import io
from os import path

from setuptools import find_packages
from setuptools import setup

here = path.abspath(path.dirname(__file__))


def _read(*names, **kwargs):
    return io.open(
        path.join(here, *names), encoding=kwargs.get("encoding", "utf8")
    ).read()


long_description = _read("README.md")

requirements = [
    "compas_fab ~= 0.10.2",
    "compas_rrc ~= 0.2.2",
    "questionary ~= 1.5.1",
    "colorama ~= 0.4.3",
    "confuse ~= 1.0.0",
    "prompt-toolkit ~= 3.0.3",
    "Pygments ~= 2.5.2",
]

extras_require = {
    "dev": [
        "attrs ~= 19.3",
        "black ~= 19.10b0",
        "doc8",
        "flake8",
        "invoke >= 0.14",
        "isort",
        "pydocstyle",
        "pytest >= 3.2",
        "recommonmark >=0.6",
        "sphinx_compas_theme >= 0.4",
        "sphinx > =1.6",
        'sphinx-autodoc-typehints[type_comments] >= 1.10 ; python_version > "2.7"',
        "setuptools_scm[toml]",
    ]
}

setup(
    name="compas_rcf",
    description="python module for MAS DFAB project Rapid Clay Formations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tetov/compas_rcf",
    author="Anton T Johansson",
    author_email="anton@tetov.se",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: IronPython",
    ],
    keywords=["architecture", "engineering", "fabrication", "construction"],
    project_urls={
        "Repository": "https://github.com/compas-dev/compas",
        "Issues": "https://github.com/compas-dev/compas/issues",
        "Documentation": "https://compas_rcf.tetov.se/",
    },
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    install_requires=requirements,
    extras_require=extras_require,
    python_requires=">=3.7",  # usage in IronPython is supported, see note in README
    entry_points={"console_scripts": []},
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import math
from itertools import count
import collections

import compas.geometry as cg
from compas import IPY
from compas.datastructures import Mesh as cg_Mesh
from compas.geometry import Frame
from compas.geometry import Translation
from compas_ghpython.artists import MeshArtist

from compas_rcf.utils.compas_to_rhino import cgframe_to_rgplane
from compas_rcf.utils.util_funcs import ensure_frame
from compas_rcf.utils.util_funcs import get_offset_frame
from compas_rcf.utils.util_funcs import list_elem_w_index_wrap

if IPY:
    import Rhino.Geometry as rg

# __all__ = ["ClayBullet", "check_id_collision", "ClayBulletEncoder"]


class ClayBullet(object):
    """Simple Clay Cylinder.

    Parameters
    ----------
    location: Rhino.Geometry.Plane, compas.geometry.Plane, compas.geometry.Frame

    radius : float, optional
        The radius of the initial cylinder.
    height : float, optional
        The height of the initial cylinder.
    compression_ratio : float (>0, <=1), optional
        The ratio of compression applied to the initial cylinder.
    """

    # creates id-s for objects
    _ids = count(0)

    def __init__(
        self,
        location,
        trajectory_to=[],
        trajectory_from=[],
        id=None,
        radius=45,
        height=100,
        compression_ratio=0.5,
        clay_density=2.0,
        precision=5,
        tool=None,
        vkey=None,
        **kwargs
    ):
        self.location = location
        self.trajectory_to = trajectory_to
        self.trajectory_from = trajectory_from

        # sortable ID, used for fabrication sequence
        if id is None:
            self.id = next(self._ids)
        else:
            self.id = str(id) + "x"  # To avoid id collisions

        self.radius = radius
        self.height = height
        self.compression_ratio = compression_ratio
        self.tool = tool
        self.vkey = vkey
        if kwargs:
            self.attributes = kwargs

        self.cycle_time = None
        self.placed = None

    @property
    def location(self):
        """Frame specifying location of bottom of clay cylinder.

        Returns
        -------
        :class:`compas.geometry.Frame`
        """
        return self._location

    @location.setter
    def location(self, frame_like):
        """Ensure that location is stored as compas.geometry.Frame object."""
        self._location = ensure_frame(frame_like)

    @property
    def placement_frame(self):
        """Last frame in placement procedure.

        Derived from location, height and compression ratio.

        Returns
        -------
        :class:`compas.geometry.Frame`
        """
        vector = self.location.zaxis * self.compressed_height * -1
        T = Translation(vector)

        return self._location.transformed(T)

    @property
    def trajectory_to(self):
        """Frames describing trajectory from picking station to placement moves.

        Returns
        -------
        list of :class:`compas.geometry.Frame`
        """
        return self._trajectory_to

    @trajectory_to.setter
    def trajectory_to(self, frame_list):
        """Ensure that trajectory_to are stored as compas.geometry.Frame objects."""
        self._trajectory_to = []

        if isinstance(frame_list, collections.Sequence):
            for frame_like in frame_list:
                frame = ensure_frame(frame_like)
                self._trajectory_to.append(frame)
        else:
            frame = ensure_frame(frame_list)
            self._trajectory_to.append(frame)

    @property
    def trajectory_from(self):
        """Frames describing trajectory from last placement move to picking station.

        Returns
        -------
        list of :class:`compas.geometry.Frame`
        """
        return self._trajectory_from

    @trajectory_from.setter
    def trajectory_from(self, frame_list):
        """Ensure that trajectory_from are stored as compas.geometry.Frame objects."""
        self._trajectory_from = []
        if isinstance(frame_list, collections.Sequence):
            for frame_like in frame_list:
                frame = ensure_frame(frame_like)
                self._trajectory_from.append(frame)
        else:
            frame = ensure_frame(frame_list)
            self._trajectory_from.append(frame)

    @property
    def centroid_frame(self):
        """Get frame at middle of uncompressed bullet."""
        return get_offset_frame(self.location, self.height)

    @property
    def compressed_centroid_frame(self):
        """Get frame at middle of compressed bullet."""
        return get_offset_frame(self.location, self.compressed_height)

    @property
    def plane(self):
        """For Compatibility with older scripts."""
        return self.location_plane

    @property
    def location_plane(self):
        """Rhino.Geometry.Plane representation of location frame.

        Returns
        -------
        Rhino.Geometry.Plane
        """
        return cgframe_to_rgplane(self.location)

    @property
    def placement_plane(self):
        """Rhino.Geometry.Plane representation of placement frame.

        Returns
        -------
        Rhino.Geometry.Plane
        """
        return cgframe_to_rgplane(self.placement_frame)

    @property
    def trajectory_to_planes(self):
        """Rhino.Geometry.Plane representations of pre frames.

        Returns
        -------
        list of Rhino.Geometry.Plane
        """
        return [cgframe_to_rgplane(frame) for frame in self.trajectory_to]

    @property
    def trajectory_from_planes(self):
        """Rhino.Geometry.Plane representation of pre frames.

        Returns
        -------
        list of Rhino.Geometry.Plane
        """
        return [cgframe_to_rgplane(frame) for frame in self.trajectory_from]

    @property
    def centroid_plane(self):
        """Get plane at middle of uncompressed bullet.

        Returns
        -------
        Rhino.Geometry.Plane
        """
        return cgframe_to_rgplane(self.centroid_frame)

    @property
    def compressed_centroid_plane(self):
        """Get plane at middle of compressed bullet.

        Returns
        -------
        Rhino.Geometry.Plane
        """
        return cgframe_to_rgplane(self.centroid_frame)

    @property
    def volume(self):
        """Volume of clay bullet in mm^3.

        Returns
        -------
        float
        """
        return math.pi * self.radius ** 2 * self.height

    @property
    def volume_m3(self):
        """Volume of clay bullet in m^3.

        Returns
        -------
        float
        """
        return self.volume * 1e-9

    @property
    def compressed_radius(self):
        """Radius of clay bullet in mm when compressed to defined compression ratio.

        Returns
        -------
        float
        """
        return math.sqrt(self.volume / (self.compressed_height * math.pi))

    @property
    def compressed_height(self):
        """Height of clay bullet in mm when compressed to defined compression ratio.

        Returns
        -------
        float
        """
        return self.height * self.compression_ratio

    @property
    def circle(self):
        """Rhino.Geometry.Circle representing bullet footprint.

        Returns
        -------
        Rhino.Geometry.Circle
        """
        return rg.Circle(self.location_plane, self.compressed_radius)

    @property
    def cylinder(self):
        """Rhino.Geometry.Cylinder representing bullet.

        Returns
        -------
        Rhino.Geometry.Cylinder
        """
        return rg.Cylinder(self.circle, self.compressed_height)

    @property
    def vector_from_bullet_zaxis(self):
        """Vector through center of bullet.

        Returns
        -------
        :class:`compas.geometry.Vector`
        """
        return self.location.normal * self.compressed_height

    @classmethod
    def from_data(cls, data):
        """Construct a :class:`ClayBullet` instance from its data representation.

        Parameters
        ----------
        data : :obj:`dict`
            The data dictionary.

        Returns
        -------
        :class:`ClayBullet`
            The constructed ClayBullet instance
        """
        location = Frame.from_data(data.pop("_location"))

        trajectory_to = []
        for frame_data in data.pop("_trajectory_to"):
            trajectory_to.append(Frame.from_data(frame_data))

        trajectory_from = []
        for frame_data in data.pop("_trajectory_from"):
            trajectory_from.append(Frame.from_data(frame_data))

        # Take the rest of the dictionary
        kwargs = data

        return cls(
            location,
            trajectory_to=trajectory_to,
            trajectory_from=trajectory_from,
            **kwargs
        )

        @classmethod
        def from_compressed_centroid_frame_like(
            cls, centroid_frame_like, compression_ratio=0.5, height=100, kwargs={}
        ):
            """Construct a :class:`ClayBullet` instance from centroid plane.

            Parameters
            ----------
            centroid_frame_like : compas.geometry.Frame or Rhino.Geometry.Plane
                frame between bottom of clay bullet and compressed top
            compression_ratio : float
                The compressed height as a percentage of the original height
            height : float
                The height of the bullet before compression
            kwargs
                Other attributes

            Returns
            -------
            :class:`ClayBullet`
                The constructed ClayBullet instance
            """
            centroid_frame = ensure_frame(centroid_frame_like)
            compressed_height = height * compression_ratio
            location = get_offset_frame(centroid_frame, -compressed_height / 2)

            return cls(
                location, compression_ratio=compression_ratio, height=height, **kwargs
            )

    def generate_mesh(self, face_count=18):
        """Generate mesh representation of bullet with custom resolution.

        Parameters
        ----------
        face_count : int, optional
            Desired number of faces, by default 18
            Used as a guide for the resolution of the mesh cylinder

        Returns
        -------
        Rhino.Geometry.Mesh
        """
        if face_count < 6:
            sides = 3
        elif face_count < 15:
            sides = 4
        else:
            sides = face_count // 3

        polygons = []
        polygons.append(rg.Polyline.CreateInscribedPolygon(self.circle, sides))

        T = rg.Transform.Translation(self.circle.Normal * -self.compressed_height)

        second_polygon = polygons[0].Duplicate()
        second_polygon.Transform(T)

        polygons.append(second_polygon)

        mesh = cg_Mesh()
        outer_verts_polygons = []

        # generate verts at polygon corners
        for polygon in polygons:
            _temp_list = []

            polygon_corners = list(polygon.Item)
            polygon_corners.pop()  # remove end pt since == start pt

            for pt in polygon_corners:
                _temp_list.append(mesh.add_vertex(x=pt.X, y=pt.Y, z=pt.Z))
            outer_verts_polygons.append(_temp_list)

        polygon_faces = []
        for vkeys in outer_verts_polygons:
            polygon_faces.append(mesh.add_face(vkeys))

        # if >4 sides polygon, create faces by tri subd
        if sides > 4:

            centroid_verts = []
            for fkey in polygon_faces:
                x, y, z = mesh.face_centroid(fkey)
                centroid_verts.append(mesh.add_vertex(x=x, y=y, z=z))
                mesh.delete_face(fkey)

            # create new faces
            for vkeys, ckey in zip(outer_verts_polygons, centroid_verts):
                for i, vkey in enumerate(vkeys):
                    next_vkey = list_elem_w_index_wrap(vkeys, i + 1)
                    mesh.add_face([ckey, vkey, next_vkey])

        # generate faces between polygons
        vertex_for_vertex = zip(*outer_verts_polygons)

        for i, mirror_corners_1 in enumerate(vertex_for_vertex):
            mirror_corners_2 = list_elem_w_index_wrap(vertex_for_vertex, i + 1)
            mesh.add_face(mirror_corners_1 + mirror_corners_2[::-1])

        # to Rhino.Geometry and clean it up
        rg_mesh = MeshArtist(mesh).draw_mesh()
        rg_mesh.UnifyNormals()
        rg_mesh.Normals.ComputeNormals()

        return rg_mesh


def check_id_collision(clay_bullets):
    """Check for duplicate ids in list of ClayBullet instances.

    Parameters
    ----------
    clay_bullets : list of :class:`ClayBullet`

    Raises
    ------
    Exception
        Raises exception when first duplicate is found
    """
    ids = [bullet.id for bullet in clay_bullets]

    set_of_ids = set()
    for id in ids:
        if id in set_of_ids:
            raise Exception(
                "Id {} appears more than once in list of ClayBullet instances".format(
                    id
                )
            )
        set_of_ids.add(id)


class ClayBulletEncoder(json.JSONEncoder):
    """JSON encoder for :class:`ClayBullet`.

    Implemented from https://docs.python.org/3/library/json.html#json.JSONEncoder
    """

    def default(self, obj):
        if isinstance(obj, ClayBullet):
            return obj.__dict__
        if isinstance(obj, cg.Primitive):
            return obj.to_data()
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_rcf.abb import *  # noqa E403
from compas_rcf.fabrication import *  # noqa E403
from compas_rcf.ur import *  # noqa E403
from compas_rcf.utils import *  # noqa E403


def test_nothing():
    assert 5 == 5



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import compas.geometry as cg
import Rhino.Geometry as rg

from compas_rcf.ur.urscript_wrapper import set_DO
from compas_rcf.ur.urscript_wrapper import set_standard_analog_out
from compas_rcf.ur.urscript_wrapper import textmsg
from compas_rcf.utils.rhino_to_compas import rgtransform_to_matrix

# Programs


def turn_off_outputs_program():
    script = ""
    script += "def program():\n"
    script += textmsg("Turning of all DOs and AOs.")
    for i in range(7):
        script += set_DO(i, False)
    for i in range(1):
        script += set_standard_analog_out(i, 0)
    script += "end\n"
    script += "program()\n"
    return script


# Geometry


def axis_angle_vector_from_plane_to_plane(plane_to, plane_from=rg.Plane.WorldXY):
    T = rg.Transform.PlaneToPlane(plane_from, plane_to)
    M = rgtransform_to_matrix(T)
    return cg.axis_angle_vector_from_matrix(M)
